#%% Center 

import numpy as np

def center(X):
    '''
    
    Center: center data
    [X_auto, mX] = center(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_auto [n x k] 
        preprocessed spectra
    mX [1 x k] 
        mean of all variables     
    
    '''
    
    m,n = X.shape
    mx = np.nanmean(X, axis=0)
    ax = (X - mx)
    
    mx = mx[np.newaxis]
    
    return ax, mx


# Colorspectra: plot X spectra with colorcode defined by y

import numpy as np
import matplotlib.pyplot as plt

def colorspectra(X,y,x_label='wavelength',y_label='intensity',colorbar_label='concentration'):
    '''
    
    Colorspectra: plot X spectra with a colorcode defined by y
    colorspectra(X,y,x_label,y_label,colorbar_label)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    Y [n x 1] <numpy.ndarray> 
        spectra
    x_label, y_label, colorbar_label [] <str> 
        optional inputs
        labels of x-axis, y-axis and colorbar
    
    OUTPUT
    none        
    
    '''

    # Plot - color by lines
    [s0,s1]=X.shape
    # Add column listing the original order of y
    order = np.arange(0,s0,1)
    order = (order[np.newaxis]).T
    color = plt.cm.inferno(np.linspace(0,0.9,s0))
    ycolor = np.concatenate((order,y),axis=1)
    # Sort "y" by increasing value
    ycolor=ycolor[np.argsort(ycolor[:,1])]
    # Add colormap
    ycolor = np.concatenate((ycolor,color),axis=1)
    # Sort "order" by increasing value to return to original order
    ycolor=ycolor[np.argsort(ycolor[:,0])]
    # Now the colormap is sorter according to y
    
    # Using contourf to provide my colorbar info, then clearing the figure
    plt.figure()
    mymap = plt.cm.inferno
    Z = [[0,0],[0,0]]
    ymin, ymax = (np.floor(np.min(y)), np.ceil(np.max(y)))
    ymin = int(ymin)
    ymax = int(ymax) 
    levels = np.arange(ymin,ymax,(ymax-ymin)/100)
    CS3 = plt.contourf(Z, levels,cmap=mymap)
    plt.close()
        
    fig = plt.figure()
    for i in range(s0):
        x = X[i,:]
        plt.plot(x.T, color=color[i])
    cbar = plt.colorbar(CS3)
    cbar.set_label(colorbar_label, rotation=270)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.xlabel(x_label,fontsize=16)
    plt.ylabel(y_label,fontsize=16)
    plt.show()
    count = int(np.round(1000*np.random.rand(1),0))
    plt.tight_layout()
    #name = 'plot colors'+str(count)+'.png'
    #name = str(name)
    #print(name)
    #fig.savefig('plot colors'+str(count)+'.png',dpi=200)


# Autoscale: center and scale to unit variance

import numpy as np

def autoscale(X):
    '''
    
    Autoscale: center and scale to unit variance
    [X_auto, mX, sX] = autoscale(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_auto [n x k] 
        preprocessed spectra
    mX [1 x k] 
        mean of all variables
    sX [1 x k] 
        standard deviation of all variables        
    
    '''
    
    m,n = X.shape
    mx = np.nanmean(X, axis=0)
    stdx = np.nanstd(X, axis=0)
    ax = (X - mx)/stdx
    
    # Remove NaN values occuring when std = 0
    for i in range(1,m):
        for j in range(1,n):
            if np.isnan(ax[i,j]) == 1:
                ax[i,j] = 0
    
    mx = mx[np.newaxis]
    stdx = stdx[np.newaxis]
    
    return ax, mx, stdx


# Linear baseline removal (remove the mean of each spectrum)

import numpy as np

def baseline(spectra):
    '''
    
    Linear baseline removal (remove the mean of each spectrum)
    [X_baseline] = baseline(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_baseline [n x k] 
        preprocessed spectra       
    
    '''
    
    spectra = spectra.T
    spectra = spectra - np.mean(spectra, axis=0)
    spectra = spectra.T
    return spectra


# Detrend: Perform spectral detrending to remove linear trend from data

import scipy

def detrend(spectra):
    '''
    
    Detrend: Perform spectral detrending to remove linear trend from data
    [X_detrend] = detrend(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_detrend [n x k] 
        preprocessed spectra      
    
    '''
    
    return scipy.signal.detrend(spectra, bp=0)


from RG.xlsread import *
from RG.regress import *
from RG.R2 import *

from RG.PCA import *
from RG.PLS import *
from RG.mbPLS import *
from RG.LDA import *


from RG.colorspectra import *

from RG.center import *
from RG.autoscale import *
from RG.baseline import *
from RG.SNV import *
from RG.MSC import *
from RG.EMSC import *
from RG.savitzky_golay import *
from RG.detrend import *













# MLR regression

import numpy as np
import statsmodels.api as sm
import pandas as pd

def regress(y, X):

    model = sm.OLS(y, X).fit()
    b = model.params
    b_int = model.conf_int(0.05)

    print('betas and confidence intervals')
        
    out = pd.DataFrame({'low': b_int[:,0], 'b': b, 'high': b_int[:,1]})
    print(out)
    
    b = (b[np.newaxis]).T # Vector format for output
    
    R2 = model.rsquared
    F = model.fvalue
    p = model.f_pvalue
    
    return b, b_int, model, R2, F, p


# Calculate R2 for MLR model

import numpy as np

def R2(y,yhat):
    # R2 in calibration
    n = y.shape[0]
    ym = np.ones((n,1))*np.mean(y) # Vector form
    SSr = np.sum((yhat-ym)**2)
    SSe = np.sum((y-yhat)**2)
    SSt = sum((y-ym)**2)
    
    R2 = 1 - SSe / SSt
    
    return R2


# MSC: Multiplicative scatter correction
# https://nirpyresearch.com/two-scatter-correction-techniques-nir-spectroscopy-python/

import numpy as np

def MSC(input_data):
    ''' 
    MSC: Multiplicative scatter correction
    [X_msc,reference] = msc(X,reference)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_msc [n x k] 
        preprocessed spectra
    
    '''
    
    
    
    # mean centre correction
    for i in range(input_data.shape[0]):
        input_data[i,:] -= input_data[i,:].mean()
    
    # Get the reference spectrum : calculate mean
    ref = np.mean(input_data, axis=0)    
        
    # Define a new array and populate it with the corrected data    
    data_msc = np.zeros_like(input_data)
    for i in range(input_data.shape[0]):
        # Run regression
        fit = np.polyfit(ref, input_data[i,:], 1, full=True)
        # Apply correction
        data_msc[i,:] = (input_data[i,:] - fit[0][1]) / fit[0][0] 
    
    ref = ref[np.newaxis]
    return (data_msc,ref) 



# Standard Normal Variate

import numpy as np
from numpy import matlib as mb

def SNV(x):
    ''' 
    
    SNV: standard normal variate
    [X_snv] = snv(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_snv [n x k] 
        preprocessed spectra
    
    '''
    
    [m,n]=x.shape
    rmean=np.mean(x,1)
    rmean = (rmean[np.newaxis]).T
    dr=x-mb.repmat(rmean,1,n)
    drsum = np.sqrt(np.sum(dr**2,1)/(n-1))
    drsum = (drsum[np.newaxis]).T
    x_snv=dr/np.matlib.repmat(drsum,1,n)
    
    return x_snv


# Savitzky Golay


def savitzky_golay(Y, window_size, order, deriv, rate=1):
    ''' 
    
    Savitzky Golay smoothing and derivative of spectral data
    [X_sg] = savitzky_golay(X,window_size, order, deriv)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    window_size [1 x 1] <numpy.ndarray>
        width of the smoothing window
    order [1 x 1] <numpy.ndarray>
        order of the polynomial function used for the smoothing
    deriv [1 x 1] <numpy.ndarray>
        degree of the derivative
  
    OUTPUT
    X_sg [n x k] 
        preprocessed spectra
    
    '''
    for i in range(Y.shape[0]):
        Y[i,:] = sav_gol(Y[i,:], window_size, order, deriv)
        
    return Y



# Sav Gol for a single spectrum
def sav_gol(y, window_size, order, deriv, rate=1):
    r"""Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    import numpy as np
    from math import factorial

    #try:
    #    window_size = np.abs(np.int(window_size))
    #    order = np.abs(np.int(order))
    #except ValueError, msg:
    #    raise ValueError("window_size and order have to be of type int")
    #if window_size % 2 != 1 or window_size < 1:
    #    raise TypeError("window_size size must be a positive odd number")
    #if window_size < order + 2:
    #    raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m[::-1], y, mode='valid')


import setuptools


setuptools.setup(
    name="RG",
    version="0.0.28",
    author="Ryan Gosselin",
    author_email="ryan.gosselin@usherbrooke.ca",
    url="https://www.usherbrooke.ca/gchimiquebiotech/departement/professeurs/ryan-gosselin/",
    packages=["RG"],
    description="Ryan's go-to Python functions",
    long_description="Miscellaneous functions:\
    \n\
    \nWork with data:\
    \n\nxlsread, regress, colorspectra, R2\
    \n\
    \nMultivariate data analysis\
    \nPCA, PLS, mbPLS, LDA\
    \n\
    \nSpectral pretreatments\
    \ncenter, autoscale, SNV, MSC, EMSC, Savitzky Golay, detrend, baseline",
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


# Read Excel documents like Matlab

#wb = 'file.xlsx' # workbook
#ws = 'sheet1' # Pick worksheet from workbook
#cell_tl = 'a20' # Cell - top left
#cell_br = 'c30' # Cell - bottom right

# Ryan Gosselin

import numpy as np
import openpyxl

def xlsread(wb, ws, cell_tl, cell_br):

    wb = openpyxl.load_workbook(wb)
    ws = wb[ws] # Pick worksheet from workbook
    ws = ws[cell_tl:cell_br] 
    
    X = []
    for row in ws:
        for obj in row:
            #print(obj.coordinate, obj.value)
            X.append((obj.value))
    col_count = len(row)
    row_count = len(ws)
    X = np.reshape(X,(row_count,col_count))

    return X


# EMSC: extended multiplicative scatter correction to the mean

import numpy as np


def EMSC(spectra):
    '''
    
    EMSC: extended multiplicative scatter correction to the mean
    [X_emsc] = EMSC(X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_emsc [n x k] 
        preprocessed spectra
       
    '''

    spectra = spectra.T
        
    wave = np.arange(0,spectra.shape[0],1)    
    p1 = .5 * (wave[0] + wave[-1])
    p2 = 2 / (wave[0] - wave[-1])

    # Compute model terms
    model = np.ones((wave.size, 4))
    model[:, 1] = p2 * (wave[0] - wave) - 1
    model[:, 2] = (p2 ** 2) * ((wave - p1) ** 2)
    model[:, 3] = np.mean(spectra, axis=1)

    # Solve correction parameters
    params = np.linalg.lstsq(model, spectra,rcond=None)[0].T

    # Apply correction
    spectra = spectra - np.dot(params[:, :-1], model[:, :-1].T).T
    spectra = np.multiply(spectra, 1 / np.repeat(params[:, -1].reshape(1, -1), spectra.shape[0], axis=0))

    spectra = spectra.T
    return spectra


# mbPLS: multiblock PLS

import numpy as np

def mbPLS(nbPC, Y, *args):
    '''
    
    HELP NOT FINISHED!!!!!
    
    mbPLS: multiblock PLS
    [X_auto, mX, sX] = mbPLS(nbPC, Y, X1, X2, Xi, ...)
    
    INPUT
    npPC [1 x 1] <numpy.ndarray>
        number of components in the mbPLS decomposition
    Y [n x m] <numpy.ndarray>
        responses
        n samples
        m variables
    Xi [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    X_auto [n x k] 
        preprocessed spectra
    mX [1 x k] 
        mean of all variables
    sX [1 x k] 
        standard deviation of all variables        
    
    '''
    
    
    
    
    
    
#%% Load data
    # Tuple containing n block matrices
    varargin = args
    n_block = len(varargin)
    
    # Creation matrix X containing n blocks    
    X = []
    for i in range(n_block):
        xi = varargin[i]
        if (i==0):
            X = xi
        else:
            X = np.append(X,xi,axis=1)

    n,k = X.shape
    n,l = Y.shape
    
    # Save copy
    varargin0 = varargin
    X0 = X
    Y0 = Y
    
    #%%  Initialization
    T = np.zeros((n,n_block))
    Tt = np.zeros((n,nbPC))
    U = np.zeros((n,nbPC))
    W = np.zeros((n_block,nbPC))
    
    X_all = []
    
    # For creation of tuple containing T for each block
    T_all = []
    T_all_temp = []
    for i in range(n_block):
        Ti = np.zeros((n,nbPC))
        T_all.append(Ti)
        
    # For creation of tuple containing W for each block
    Wi_all = []
    Wi_all_temp = []
    for i in range(n_block):
        Xi = varargin[i]
        n,ki = Xi.shape
        Wi = np.zeros((ki,nbPC))
        Wi_all.append(Wi)
           
   # Sum of squares
    SSXi = np.zeros((nbPC,n_block))
    SSY = np.zeros((nbPC,1))

    
    # %% PLS
    for a in range(nbPC):
        u = np.ones((n,1))
        u_temp = np.ones((n,1))
        tt_temp = np.ones((n,1))
        error_u = 1
        error_tt = 1
        
        while (error_u > 1E-20) and (error_tt > 1E-20):
            
                    for i in range(n_block):
                        Xi = varargin[i]
                        wi = (Xi.T@u)/(u.T@u)
                        wi = wi / np.linalg.norm(wi)
                        ti = Xi@wi
                        
                        T[:,i] = np.squeeze(ti)
                      
                        Wi = Wi_all[i]
                        Wi[:,a] = np.squeeze(wi)
                        Wi_all_temp.append(Wi)
                    Wi_all = Wi_all_temp
                    Wi_all_temp = []
                    
                    wt = (T.T@u)/(u.T@u)
                    wt = wt/np.linalg.norm(wt)
                    tt = (T@wt)/(wt.T@wt)
        
                    q = (Y.T@tt) / (tt.T@tt)
                    u = (Y@q) / (q.T@q)
        
                    error_tt = sum(np.power((tt-tt_temp),2),0) 
                    tt_temp = tt
        
                    error_u = sum(np.power((u-u_temp),2),0) 
                    u_temp = u
                                       
                    
        for i in range(n_block):
            # Deflation 
            Xi = varargin[i]
            pi = (Xi.T@tt)/(tt.T@tt) 
            Xi = Xi - tt@pi.T
            X_all.append(Xi)
            
            # Sum of squares for X
            Xi_hat = tt@pi.T
            Xi_0 = varargin0[i]
            ssXi_0 = np.sum(np.sum(Xi_0*Xi_0))
            ssXi = np.sum(np.sum(Xi_hat*Xi_hat))
            ssXi = ssXi/ssXi_0
            
            SSXi[a,i] = ssXi
            
            
        # Update of deflated blocks
        varargin = X_all
        X_all = []
        
        pt = (X.T@tt)/(tt.T@tt)
    
        # Scaling based on PLS Toolbox
#        tt = tt*np.linalg.norm(pt)  
#        wt = wt*np.linalg.norm(pt)
#        pt = pt/np.linalg.norm(pt)
            
        b = (u.T@tt)/(tt.T@tt)
        Y = Y - b*tt@q.T 
        
        # Sum of squares for Y
        Yhat = b*tt@q.T
        ssY0 = np.sum(np.sum(Y0*Y0))
        ssY = np.sum(np.sum(Yhat*Yhat))
        ssY = ssY / ssY0;
        SSY[a] = ssY
            
        
        # Save 
        Tt[:,a] = np.squeeze(tt)
        U[:,a] = np.squeeze(u)
        W[:,a] = np.squeeze(wt)
        
        for i in range(n_block):
            Ti = T_all[i]
            Ti[:,a] = T[:,i]
            T_all_temp.append(Ti)
        T_all = T_all_temp
        T_all_temp = []
    
                          
    SSX = np.sum(SSXi,axis=1)/n_block
 
    return Tt, T_all, W, Wi_all, SSX, SSY, SSXi
    
        